@@ -1,5 +1,6 @@
 /* Double.java -- object wrapper for double
-   Copyright (C) 1998, 1999, 2000, 2001, 2002 Free Software Foundation, Inc.
+   Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2005, 2006
+   Free Software Foundation, Inc.
 
 This file is part of GNU Classpath.
 
@@ -15,8 +16,8 @@ General Public License for more details.
 
 You should have received a copy of the GNU General Public License
 along with GNU Classpath; see the file COPYING.  If not, write to the
-Free Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
-02111-1307 USA.
+Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+02110-1301 USA.
 
 Linking this library statically or dynamically with other modules is
 making a combined work based on this library.  Thus, the terms and
@@ -35,10 +36,8 @@ this exception to your version of the library, but you are not
 obligated to do so.  If you do not wish to do so, delete this
 exception statement from your version. */
 
-
 package java.lang;
 
-import gnu.classpath.Configuration;
 
 /**
  * Instances of class <code>Double</code> represent primitive
@@ -48,8 +47,8 @@ import gnu.classpath.Configuration;
  * related to doubles.
  *
  * @author Paul Fisher
- * @author Andrew Haley <aph@cygnus.com>
- * @author Eric Blake <ebb9@email.byu.edu>
+ * @author Andrew Haley (aph@cygnus.com)
+ * @author Eric Blake (ebb9@email.byu.edu)
  * @since 1.0
  * @status updated to 1.4
  */
@@ -88,6 +87,12 @@ public final class Double extends Number implements Comparable
    */
   public static final double NaN = 0.0 / 0.0;
 
+  /**
+   * The number of bits needed to represent a <code>double</code>.
+   * @since 1.5
+   */
+  public static final int SIZE = 64;
+
   /**
    * The primitive type <code>double</code> is represented by this
    * <code>Class</code> object.
@@ -102,18 +107,6 @@ public final class Double extends Number implements Comparable
    */
   private final double value;
 
-  /**
-   * Load native routines necessary for this class.
-   */
-  static
-  {
-    if (Configuration.INIT_LOAD_LIBRARY)
-      {
-	System.loadLibrary("javalang");
-	initIDs();
-      }
-  }
-
   /**
    * Create a <code>Double</code> from the primitive <code>double</code>
    * specified.
@@ -179,6 +172,22 @@ public final class Double extends Number implements Comparable
     return toString(d, false);
   }
 
+  /**
+   * Returns a <code>Double</code> object wrapping the value.
+   * In contrast to the <code>Double</code> constructor, this method
+   * may cache some values.  It is used by boxing conversion.
+   *
+   * @param val the value to wrap
+   * @return the <code>Double</code>
+   * 
+   * @since 1.5
+   */
+  public static Double valueOf(double val)
+  {
+    // We don't actually cache, but we could.
+    return new Double(val);
+  }
+
   /**
    * Create a new <code>Double</code> object using the <code>String</code>.
    *
@@ -243,7 +252,7 @@ public final class Double extends Number implements Comparable
    * @see #NEGATIVE_INFINITY
    * @since 1.2
    */
-  public static native double parseDouble(String s);
+  public static native double parseDouble(String str);
 
   /**
    * Return <code>true</code> if the <code>double</code> has the same
@@ -392,9 +401,9 @@ public final class Double extends Number implements Comparable
    * <code>Double.NaN</code> as equal, but treats <code>0.0</code> and
    * <code>-0.0</code> as unequal.
    *
-   * <p>Note that <code>d1.equals(d2)<code> is identical to
+   * <p>Note that <code>d1.equals(d2)</code> is identical to
    * <code>doubleToLongBits(d1.doubleValue()) ==
-   *    doubleToLongBits(d2.doubleValue())<code>.
+   *    doubleToLongBits(d2.doubleValue())</code>.
    *
    * @param obj the object to compare
    * @return whether the objects are semantically equal
@@ -428,7 +437,9 @@ public final class Double extends Number implements Comparable
    * @return the bits of the <code>double</code>
    * @see #longBitsToDouble(long)
    */
+  // GCJ LOCAL: We diverge from Classpath for efficiency.
   public static native long doubleToLongBits(double value);
+  // END GCJ LOCAL
 
   /**
    * Convert the double to the IEEE 754 floating-point "double format" bit
@@ -444,7 +455,9 @@ public final class Double extends Number implements Comparable
    * @return the bits of the <code>double</code>
    * @see #longBitsToDouble(long)
    */
+  // GCJ LOCAL: We diverge from Classpath for efficiency.
   public static native long doubleToRawLongBits(double value);
+  // END GCJ LOCAL
 
   /**
    * Convert the argument in IEEE 754 floating-point "double format" bit
@@ -459,7 +472,9 @@ public final class Double extends Number implements Comparable
    * @see #doubleToLongBits(double)
    * @see #doubleToRawLongBits(double)
    */
+  // GCJ LOCAL: We diverge from Classpath for efficiency.
   public static native double longBitsToDouble(long bits);
+  // END GCJ LOCAL
 
   /**
    * Compare two Doubles numerically by comparing their <code>double</code>
@@ -528,10 +543,4 @@ public final class Double extends Number implements Comparable
    */
   // Package visible for use by Float.
   static native String toString(double d, boolean isFloat);
-
-  /**
-   * Initialize JNI cache.  This method is called only by the
-   * static initializer when using JNI.
-   */
-  private static native void initIDs();
 }
